/*
 * Copyright 2010-2015 Amazon.com, Inc. or its affiliates. All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License").
 * You may not use this file except in compliance with the License.
 * A copy of the License is located at
 *
 *  http://aws.amazon.com/apache2.0
 *
 * or in the "license" file accompanying this file. This file is distributed
 * on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
 * express or implied. See the License for the specific language governing
 * permissions and limitations under the License.
 */

/**
 * @file subscribe_publish_sample.c
 * @brief simple MQTT publish and subscribe on the same topic
 *
 * This example takes the parameters from the aws_iot_config.h file and establishes a connection to the AWS IoT MQTT Platform.
 * It subscribes and publishes to the same topic - "sdkTest/sub"
 *
 * If all the certs are correct, you should see the messages received by the application in a loop.
 *
 * The application takes in the certificate path, host name , port and the number of times the publish should happen.
 *
 */

/*
    Claudio Baldini rev00.01 20/08/2020
    initial release
*/


/* Legato Framework */
#include "legato.h"
#include "interfaces.h"
#include "le_print.h"
#include "le_thread.h"
#include "le_clock.h"
#include "errno.h"
#include "le_basics.h"


/*Libraries for Embedded C SDK from AWS*/
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <unistd.h>
#include <limits.h>
#include <string.h>


uint32_t  modemProfileIndex  = 1;
le_mdc_ProfileRef_t modemProfile ;
le_mdc_ConState_t modemState;
le_timer_Ref_t   _pubTimerRef = NULL;

// Semaphore
static le_sem_Ref_t   ModemConnectSemaphore;
le_clk_Time_t   modemConnectTimeToWait;
bool modemConnecting = false;
bool modemConnected = false;

static le_result_t ConnectModem(le_mdc_ProfileRef_t profileRefPtr){
   
    //check if modem is connected now, if already connected, then skip connection
    LE_ASSERT( le_mdc_GetSessionState(modemProfile, &modemState) == LE_OK );
    if(LE_MDC_CONNECTED == modemState){
        LE_INFO("Modem is already connected, return.");
        modemConnected = true;
        modemConnecting = false;
        return LE_OK;
    }

    if(false == modemConnecting){
        modemConnecting = true; //prevent for multiple calls of function
        //le_mdc_ConnectService();
        // Get the profile reference
        LE_INFO("Modem will start a new data session.");

        //profileRefPtr = le_mdc_GetProfile(modemProfileIndex);
        //LE_ASSERT(NULL != profileRefPtr);

        //create an handler for Data Connection state
        char data[200];
        const char * APN = "telstra.wap";
        le_result_t rs;
        //stop session if already started
        LE_FATAL_IF(profileRefPtr == NULL, "Profile doesn`t exist");
        //set the PDP


        rs = le_mdc_SetPDP(profileRefPtr, LE_MDC_PDP_IPV4V6);
        LE_FATAL_IF(rs != LE_OK, "Couldn`t set PDP.");
        //readback the PDP
        le_mdc_Pdp_t ProfilePDP = le_mdc_GetPDP(profileRefPtr);
        switch (ProfilePDP)
        {   
            case LE_MDC_PDP_UNKNOWN:
                LE_INFO("PDP Unknown");
                break;
            case LE_MDC_PDP_IPV4:
                LE_INFO("PDP IPv4");
                break;
            case LE_MDC_PDP_IPV6:
                LE_INFO("IPDP Pv6");
                break;
            case LE_MDC_PDP_IPV4V6:
                LE_INFO("PDP IPv4v6");
                break;
            default:
                LE_ERROR("PDP unsupported %d", ProfilePDP);
                break;
        }
        //get APN
        if(LE_OK == le_mdc_GetAPN(profileRefPtr, &data[0], sizeof(data)))
            LE_INFO("Got APN: %s", data);
        if(0 != strcmp(APN, &data[0])){
            //set APN
            LE_INFO("Need to setup APN");
            rs = le_mdc_SetAPN(profileRefPtr,APN);
            LE_FATAL_IF(rs != LE_OK, "Couldn`t set APN.");
            rs = le_mdc_GetAPN(profileRefPtr, &data[0], sizeof(data));
            if(LE_OK == rs)
                LE_INFO("APN: %s, %d",data, rs);
        }

        LE_INFO("Try to Start Session.");
        if (LE_OK == le_mdc_StartSession(profileRefPtr))
            LE_INFO("Modem is connecting.");

        modemConnecting = false;

    }
    else
        LE_INFO("Modem is already trying to connect.");

    return LE_OK;
}

//--------------------------------------------------------------------------------------------------
/**
 *  Timer Handler
 */
//--------------------------------------------------------------------------------------------------
static void pubTimerHandler(le_timer_Ref_t  timerRef)
{
    LE_INFO("Time handler is starting");

    //check if modem is connected now.
    LE_ASSERT( le_mdc_GetSessionState(modemProfile, &modemState) == LE_OK );
    if(LE_MDC_DISCONNECTED == modemState){
        LE_INFO("Modem is disconnected, will try to re-connect");
        ConnectModem(modemProfile);
        LE_INFO("Connection function will wait for handler to return");
        if(LE_TIMEOUT == le_sem_WaitWithTimeOut(ModemConnectSemaphore, modemConnectTimeToWait))
            LE_INFO("handler hasn`t returned any value");

        return;
    }
    else
        if(LE_MDC_CONNECTED != modemState){
            LE_INFO("Modem needs to be connected to publish to AWS, function is returning.");
            return;       
        }

    LE_INFO("======Modem is connected.========");
}

//! [Sessions]
//--------------------------------------------------------------------------------------------------
/**
 * Set the network configuration.
 *
 */
//--------------------------------------------------------------------------------------------------
static void SetNetworkConfiguration(le_mdc_ProfileRef_t profileRef)
{
    char ipAddr[100] = {0};
    char gatewayAddr[100] = {0};
    char dns1Addr[100] = {0};
    char dns2Addr[100] = {0};
    char systemCmd[200] = {0};
    FILE* resolvFilePtr;
    le_mdc_ConState_t state = LE_MDC_DISCONNECTED;
    mode_t oldMask;

    // Check the state
    LE_ASSERT( le_mdc_GetSessionState(profileRef, &state) == LE_OK );
    if( state != LE_MDC_CONNECTED ){
        LE_INFO("Modem is not connected, will return...");
        return;
    }

    // Get IP, gateway and DNS addresses for IPv4 or IPv6 connectivity
    if ( le_mdc_IsIPv4(profileRef) )
    {
        LE_ASSERT(le_mdc_GetIPv4Address(profileRef, ipAddr, sizeof(ipAddr)) == LE_OK);
        LE_INFO("%s", ipAddr);

        LE_ASSERT(le_mdc_GetIPv4GatewayAddress(profileRef, gatewayAddr, sizeof(gatewayAddr))
                                                                                          == LE_OK);
        LE_INFO("%s", gatewayAddr);

        LE_ASSERT(le_mdc_GetIPv4DNSAddresses( profileRef,
                                              dns1Addr, sizeof(dns1Addr),
                                              dns2Addr, sizeof(dns2Addr)) == LE_OK );
        LE_INFO("%s", dns1Addr);
        LE_INFO("%s", dns2Addr);

        snprintf(systemCmd, sizeof(systemCmd), "/sbin/route add default gw %s", gatewayAddr);
    }
    else if ( le_mdc_IsIPv6(profileRef) )
    {
        LE_ASSERT(le_mdc_GetIPv6Address(profileRef, ipAddr, sizeof(ipAddr)) == LE_OK);
        LE_INFO("%s", ipAddr);

        LE_ASSERT(le_mdc_GetIPv6GatewayAddress(profileRef, gatewayAddr, sizeof(gatewayAddr))
                                                                                          == LE_OK);
        LE_INFO("%s", gatewayAddr);

        LE_ASSERT(le_mdc_GetIPv6DNSAddresses( profileRef,
                                              dns1Addr, sizeof(dns1Addr),
                                              dns2Addr, sizeof(dns2Addr)) == LE_OK );
        LE_INFO("%s", dns1Addr);

        snprintf(systemCmd, sizeof(systemCmd), "/sbin/route -A inet6 add default gw %s", gatewayAddr);
    }
    //LE_INFO("Sleeping for 5 secs.");
    //sleep(5);

    LE_INFO("Write routes to sbin");
    LE_INFO("%s", systemCmd);
    LE_INFO("trying to write routes: %d", system(systemCmd));
    //LE_ASSERT( system(systemCmd) == 0 );

    // allow fopen to create file with mode=644
    oldMask = umask(022);

    // open the resolver configuration
    resolvFilePtr = fopen("/etc/resolv.conf", "w+");

    if (resolvFilePtr == NULL)
    {
        LE_ERROR("Unable to open resolv.conf: %m");
    }
    LE_ASSERT( resolvFilePtr != NULL );

    LE_ASSERT( fprintf(resolvFilePtr, "nameserver %s\n", dns1Addr) > 0 );

    if (dns2Addr[0] != '\0')
    {
        LE_ASSERT( fprintf(resolvFilePtr, "nameserver %s\n", dns2Addr) > 0 );
    }

    LE_ASSERT( fclose(resolvFilePtr) == 0 );

    // restore old mask
    umask(oldMask);

}

//--------------------------------------------------------------------------------------------------
/**
 * Handler function for session state Notifications for modem connectivity
 *
 */
//--------------------------------------------------------------------------------------------------
static void StateChangeHandler(le_mdc_ProfileRef_t profileRef,le_mdc_ConState_t ConnectionStatus,void* contextPtr)
{
    LE_DEBUG("\n================================================");
    LE_ASSERT( le_mdc_GetSessionState(profileRef, &modemState) == LE_OK );
    LE_INFO("=====Handler is returning %d while profile status is %d=====", ConnectionStatus, modemState);
    static bool sessionAlreadyStopped = false;
    
    if (modemState == LE_MDC_DISCONNECTED)
    {   
            
        // Get disconnection reason
        if (LE_MDC_PDP_IPV4V6 == le_mdc_GetPDP(profileRef))
        {
            LE_INFO("Disconnection reason IPv4: %d", le_mdc_GetDisconnectionReasonExt(profileRef, LE_MDC_PDP_IPV4));
            LE_INFO("Platform specific disconnection reason IPv4: %d", le_mdc_GetPlatformSpecificDisconnectionCodeExt(profileRef,
                                                                              LE_MDC_PDP_IPV4));
            LE_INFO("Disconnection reason IPv6: %d", le_mdc_GetDisconnectionReasonExt(profileRef, LE_MDC_PDP_IPV6));
            LE_INFO("Platform specific disconnection reason IPv6: %d", le_mdc_GetPlatformSpecificDisconnectionCodeExt(profileRef,
                                                                              LE_MDC_PDP_IPV6));
        }
        else
        {
            LE_PRINT_VALUE("%d", le_mdc_GetDisconnectionReasonExt(profileRef, 0));
            LE_PRINT_VALUE("%d", le_mdc_GetPlatformSpecificDisconnectionCodeExt(profileRef, 0));
        }
        LE_INFO("=====MODEM DISCONNECTED======");
        if(false == sessionAlreadyStopped) {
            LE_INFO("Modem DISCONNECTED, will stop any ongoing communication if present");
            le_mdc_StopSession(profileRef);
            sessionAlreadyStopped = true;
            LE_INFO("sleep for 10 seconds");
            modemConnecting = false;
            sleep(10);
        }
        else
            LE_INFO("Modem already disconnected.");
    }
    else //visualize the current status
        switch(modemState)
        {
            case LE_MDC_CONNECTED:
                LE_INFO("Modem connected, will setup the network configuration.");
                SetNetworkConfiguration(profileRef);
                LE_INFO("Now release semaphore.");
                le_sem_Post(ModemConnectSemaphore); //release semaphore
                //modemConnecting = false;
                sessionAlreadyStopped = false;
                LE_INFO("sleep for 5 seconds");
                sleep(5);
                break;
            case LE_MDC_AUTHENTICATING:
                LE_INFO("Modem AUTHENTICATING, please wait...");
                break;
            case LE_MDC_SUSPENDING:
                LE_INFO("Modem SUSPENDING, please wait...");
                break;
            case LE_MDC_INCOMING:
                LE_INFO("Modem INCOMING, please wait...");
                break;
            default:
                LE_INFO("Modem state unknown.");
                break;
        }

    LE_DEBUG("\n================================================");

}

 
COMPONENT_INIT
{
    le_result_t rs = 0;

    LE_INFO("---------- Launching test modem function----------");

    //ESN=Electronic Serial number will be used together with the account name for presenting data on Grafana
    char data[200];
    LE_INFO("======== Read ESN from device  ========");
    if(LE_OK == le_info_GetEsn(&data[0], sizeof(data)))
        LE_INFO("ESN: %s, %d",data, rs);
    else
        LE_INFO("Could not read Electronic Serial Number.");

    //now start the modem
    LE_INFO("======== Checking modem configuration  ========");
    //le_mdc_ConnectService();    //connect to modem service
    modemProfile = le_mdc_GetProfile(modemProfileIndex);
    LE_INFO("Default modem profile index is: %d my profile index is %d", LE_MDC_DEFAULT_PROFILE, modemProfileIndex);
    modemState = LE_MDC_DISCONNECTED; //preset value of modemState to disconnected

    LE_INFO("Create Handler for Connection State.");
    le_mdc_SessionStateHandlerRef_t SessionStateHandler = le_mdc_AddSessionStateHandler(modemProfile, StateChangeHandler, NULL);
    LE_FATAL_IF(SessionStateHandler == NULL, "Session state handler not created!!");

    LE_ASSERT( le_mdc_GetSessionState(modemProfile, &modemState) == LE_OK );
    
    if(modemState == LE_MDC_DISCONNECTED) //only if modem is disconnected then start a new connection
    {
        modemConnected = false;
        LE_INFO("Will start connecting the modem.");
        rs = ConnectModem(modemProfile);
    }
    else 
    {
        if(modemState == LE_MDC_CONNECTED)
        {
            modemConnected = true;
            LE_INFO("Modem already connected.");
        }
    }
    

    //Connect modem and start semaphore
    ModemConnectSemaphore = le_sem_Create("ModemConnectSemaphore",0);
    modemConnectTimeToWait.sec = 3;
    modemConnectTimeToWait.usec = 0;

    //in all other cases, wait for the handler to look after the connection

    rs =  le_info_GetImei(&data[0], sizeof(data));
    LE_INFO("IMEI: %s, %d",data, rs);
    //now try to connect the router

    //Create timer to call mqttClient_ProcessEvent
    le_clk_Time_t   interval = { 3, 0 };
    le_result_t  res;

     //Create timer to publish data
    _pubTimerRef = le_timer_Create("mqttSamplePubTimer");
    LE_FATAL_IF(_pubTimerRef == NULL, "timerApp timer ref is NULL");
    res = le_timer_SetInterval(_pubTimerRef, interval);
    LE_FATAL_IF(res != LE_OK, "set interval to %lu seconds: %d", interval.sec, res);
    //set repeat times
    res = le_timer_SetRepeat(_pubTimerRef, 0);
    LE_FATAL_IF(res != LE_OK, "set repeat to indefinetely: %d", res);

    //specify the handler
    le_timer_SetHandler(_pubTimerRef, pubTimerHandler);
    //arm the timer
    le_timer_Start(_pubTimerRef);

    return;
}